/* ---------------------------------------------------------------------------
 * Tseng Labs ET6000 chipset driver
 * --------------------------------------------------------------------------
 *
 * Copyright (C) 1997 Jim Ursetto [ursetto@uiuc.edu]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ---------------------------------------------------------------------------
 * 
 * ---------------------------------------------------------------------------
 * 
 * This code was derived from the following sources of information:
 *
 * [1] Prerelease VGADOC5 (Tseng.txt/vgaregs.txt) by Finn Thoegersen
 * [2] IBM VGA driver by Jason McMullan
 * [3] XFree86 Tseng ET4000/6000 driver -- Copyright 1990,91 by Thomas
 *     Roell, Dinkelscherben, Germany
 * [4] S3 86c765 (Trio64V+) driver by Jon Taylor
 *
 */

#include <ggi/maintainers.h>
/*#define	MAINTAINER	Jim_Ursetto */
#define MAINTAINER "Jim Ursetto <ursetto@uiuc.edu>"
#define DRIVER_NAME	"Tseng Labs ET6000 chipset driver"
#define DRIVER_REV	"$Revision: 0.00 $"

#define DEBUG_LEVEL	0

#include <ggi/system.h>
#include <ggi/module.h>

#include "IBM/vga.h"
#include "Tseng/et6000.h"

#undef 	CURSOR_HW	/* Hardware cursor */

/* No longer used. */
/*#define TEXT_A000 */	/* Base of textmode framebuffer */
/*#define TEXT_LINEAR */ /* Linear framebuffer for text */

#define TEXT_LEN 0x20000

#define	POINTER_FONT	/* Font-hack mouse cursor in textmodes */
#define SLOW_POINTER	/* #define this if your pointer flickers */
/* #define DEBUG_STATE */

static struct kgi_chipset chipinfo =
{
	"Tseng Labs",			/* Manufacturer				*/
	"ET6000",			/* Model				*/
	GM_TEXT16 | GM_1BIT |	/* Available modes			*/
	GM_4BIT | GM_8BIT | GM_15BIT | GM_16BIT | GM_24BIT | GM_32BIT,
	0,			/* Flags				*/
	{ 2048, 2048 },		/* Maximum resolution			*/
	0x0,		        /* addressable RAM (autodetect)         */
        { 0000000, 135000000 }	/* DCLK limits				*/
};

static struct mem_region et6000_linear_mem =
{
             MEM_NULL,
               0x0,           /* linear framebuffer (autodetect) */
               0x0,           /* mem size (autodetect) */
               -1,            /* Decode all bits */
               "Tseng Labs ET6000 linear framebuffer"
};

static struct io_region et6000_io =
{
             IO_NULL,
             0x3C0, /* VGA IO Region Start */
             0x20, /* 32 register (0x3C0-0x3DF) */
             0x3FF, /* Only 10 address lines are decoded.. */
             "Tseng Labs ET6000 I/O region"
};    

static struct io_region et6000_pci_io = 
{
             IO_NULL,
             0,
             0x87, /* Number of PCI registers */
             -1,
             "Tseng Labs ET6000 PCI I/O region"
};

static struct mem_region et6000_mmio =
{
            MEM_NULL,
            0,        /* base ptr */
            0xFF,     /* size */
            -1,       /* Decode all bits */
            "Tseng ET6000 MMIO region"
};

static struct kgi_mmio_region et6000_linear_mmio;

/* Private data not used at this time. */
struct chipset_private
{
    struct mem_region       mem_law_region;
    struct io_region        pci_io_region;
};

#define GET_PRIVATE \
        struct chipset_private *private = KGIM_PRIVATE(dpy, chipset)

#define GET_PCI \
        struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci)

#define PCI_IO (et6000_pci_io.baseptr)
#define PCI_IO_REGION (et6000_pci_io)
#define VGA_IO (et6000_io.baseptr)
#define VGA_IO_REGION (et6000_io)
#define LINEAR_MEM_REGION (et6000_linear_mem)
#define LINEAR_MEM_SIZE (et6000_linear_mem.size)
#define LINEAR_MEM (et6000_linear_mem.baseptr)
#define LINEAR_MMIO (et6000_linear_mmio)

#include "et6000_io.inc"

/* This table holds the default values for the VGA registers.  A copy	*/
/* of it will be filled in as appropriate during mode programming in	*/
/* kgim_chipset_set_mode().						*/
static struct statetable default_state =
{
	0x23,							/* MISC	*/
	{
		0x03, 0x00, 0x00, 0x00, 0x00			/* SEQ	*/
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* CRT	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x1x	*/
		0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC	*/
		0xFF
	},
	{
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC	*/
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x00, 0x00, 0x0F, 0x00, 0x00
	},
        0x00                                                    /* VidCtrl1 */
};


struct statetable reset_state;
ggi_mode default_ggi_mode;
struct kgi_mode default_kgi_mode;

static uint32 calc_mem(uint8 bus, uint8 device_fn);

/* Warning: the PCI detection code in here is Linux-specific and must be changed to
 * use GGI abstract PCI functions. */

static int verify_et6000(void)
{
    uint16 status;
    uint8 result8;
    uint32  result32;
    uint32 et6kbase;
    uint8 bus, device_fn;
    
    status = pcibios_find_device(PCI_VENDOR_ID_TSENG,
				 PCI_DEVICE_ID_TSENG_ET6000,
				 0, &bus, &device_fn);

    
    if (status != PCIBIOS_SUCCESSFUL)
    {
	ERROR("ET6000 not detected.\n");
	return -E(CHIPSET, UNKNOWN);
    }

    outbMISC(inb(MISCr));	/* Adjust IO base       */
    
    pcibios_read_config_byte(bus, device_fn, PCI_REVISION_ID, &result8);
    if (result8 < 70)
        DEBUG1("Tseng Labs ET6000 (rev. %d) found.\n", result8);
    else 
    {
        ERROR("Tseng Labs ET6100 found but not supported yet.\n");
	return -ENODEV;
    }
    
    pcibios_read_config_dword(bus, device_fn, PCI_BASE_ADDRESS_0, &result32);
    result32 &= 0xff000000;
    DEBUG1("Linear framebuffer found at 0x%08x.\n", result32);
    pcibios_read_config_dword(bus, device_fn, PCI_BASE_ADDRESS_1, &et6kbase);
    et6kbase &= ~0xff;
    DEBUG1("PCI I/O is mapped at port 0x%04x.\n", et6kbase);

    PCI_IO_REGION.base = et6kbase;
    LINEAR_MEM_REGION.base = result32;

    chipinfo.ramsize = calc_mem(bus, device_fn);
    if (!chipinfo.ramsize)
        return -E(CHIPSET, UNKNOWN);

    et6000_linear_mem.size = chipinfo.ramsize;  /* this'll have to be decreased for accel or hw cursor */
    et6000_linear_mmio.next	= NULL;
    et6000_linear_mmio.name	= "Video RAM";
    et6000_linear_mmio.type	= MMAP_TYPE_MMIO | MMAP_PUBLIC | MMAP_FRAMEBUFFER;
    et6000_linear_mmio.prot	= MMIO_RDWR | MMIO_SHARED;
    et6000_linear_mmio.psize = chipinfo.ramsize; /* warning -- autodetect needed!!! */
    et6000_linear_mmio.size =  chipinfo.ramsize; /* warning -- autodetect needed!!! */
    et6000_linear_mmio.pstart = LINEAR_MEM_REGION.base;
    return EOK;

}

/* The memory type (MDRAM, DRAM) should be stored for later use. */
/* As discussed in the XF86 driver, 2.25 meg ram is detected incorrectly;
 * this will be fixed shortly. */
static uint32 calc_mem(uint8 bus, uint8 device_fn)
{
    int type;
    uint8 size;
    uint32 mem_size = 0;
    
    enable_ext();
    type = io_in8(MISCw) & 0x03;
    disable_ext();
    pcibios_read_config_byte(bus, device_fn, PCI_TSTSETUP, &size);
    mem_size = ((size & PCI47_RFN) + 1) << 8;
    
    switch(type)
    {
      case 0x03:
	pcibios_read_config_byte(bus, device_fn, PCI_MEMCONF, &size);
	if (size & PCI45_MDRAMDBL)
	     mem_size <<= 1;
	DEBUG1("%d KB MDRAM found.\n", mem_size);
	break;
      case 0x00:
	pcibios_read_config_byte(bus, device_fn, PCI_MEMCONF, &size);
	mem_size = 1024 << (size & PCI45_DRAMSIZE);
	DEBUG1("%d KB DRAM found.\n", mem_size);
	break;
      default:
	ERROR("Video RAM detection failed!\n");
    }
    return mem_size << 10;
}

/* GGI interface */

/* Mode specific frame buffers */
/* warning: probably implies POINTER_FONT */
#include "et6000_text16.inc" 
#include "et6000_fb.inc"

static void chipset_setup_dpy(struct kgi_display *dpy, struct kgi_mode *mode)
{
	switch (mode->request.graphtype) {
		case GT_TEXT16:
		dpy->set_font		= text16_set_font;
		dpy->set_origin		= text16_set_origin;
		dpy->set_split		= text16_set_split;
		dpy->set_frame		= text16_set_frame;
		dpy->show_cursor	= text16_show_cursor;
		dpy->hide_cursor	= text16_hide_cursor;
		dpy->undo_cursor	= text16_undo_cursor;
		dpy->show_pointer	= text16_show_pointer;
		dpy->hide_pointer	= text16_hide_pointer;
		dpy->undo_pointer	= text16_undo_pointer;
		dpy->fb.dpy		= dpy;
		dpy->fb.size.x 		= mode->request.virtual.x / 
			mode->request.text.x;
		dpy->fb.size.y		= mode->request.virtual.y / 
			mode->request.text.y;
		dpy->fb.bpp		= MAP_16_BPP;
	        dpy->fb.pbuf            = LINEAR_MEM;
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= &text16_color_io;
		dpy->driver_data	=text16_data_realloc(dpy->driver_data);
		if (dpy->driver_data == NULL)
			INTERNALERROR;
		break;
		default:
		dpy->set_font		= NULL;
		dpy->set_origin		= et6000_fb_set_origin;
		dpy->set_split		= et6000_fb_set_split;
		dpy->set_frame		= et6000_fb_set_frame;
		dpy->show_cursor	= NULL;
		dpy->hide_cursor	= NULL;
		dpy->undo_cursor	= NULL;
		dpy->show_pointer	= NULL;
		dpy->hide_pointer	= NULL;
		dpy->undo_pointer	= NULL;
		dpy->fb.dpy		= dpy;
		dpy->fb.size.x		= 0;
		dpy->fb.size.y		= 0;
		dpy->fb.bpp		= 0;
		dpy->fb.pbuf		= NULL;
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= NULL;
		kgim_graphics_set_mode(dpy,mode);
	}
	memcpy(&(dpy->mode), mode, sizeof(struct kgi_mode));
}

/* Takes a mode and a bits-per-pixel value */
static void chipset_setup_mode(struct kgi_mode *mode)
{
	int bpp=0;

	/* By default, setup for graphics modes */

	switch (mode->request.graphtype) {
		case GT_TEXT16:
		mode->fb	= LINEAR_MEM;
		mode->fb_size	= TEXT_LEN;
		mode->fsize	= (mode->request.virtual.x / mode->request.text.x) *
			(mode->request.virtual.y / mode->request.text.y) * 2;
		mode->mem	= &text16_color_mem;
		mode->io	= &text16_color_io;
	        mode->mmio      = &et6000_linear_mmio;
	        return; 	/* Done setting up */
	    
		case GT_1BIT:
		bpp = 1;
		break;
		case GT_4BIT:
		bpp = 4;
		break;
		case GT_8BIT:
		bpp = 8;
		break;
	case GT_15BIT:
	case GT_16BIT:
	    bpp = 16;
	    break;
	case GT_24BIT:
	    bpp = 24;
	    break;
	case GT_32BIT:
	    bpp = 32;
	    break;
	default:
		INTERNALERROR;
		return;
	}
	
        mode->fb	= LINEAR_MEM;
        mode->fb_size	= LINEAR_MEM_SIZE;
	mode->fsize	= (mode->request.virtual.x * 
			   mode->request.virtual.y * 8) / bpp;

        DEBUG2("requested x: %d y: %d vx: %d vy: %d\n", mode->request.visible.x, mode->request.visible.y,
	                                                mode->request.virtual.x, mode->request.virtual.y);
	mode->mem	= NULL;
	mode->io	= NULL;
        /* Perhaps here I can set mmio to the 4-bit one, but I don't think it makes sense in linear mode. */
        mode->mmio      = &et6000_linear_mmio;
}

/* Is this ever called? */
static int chipset_get_reset_mode(struct kgi_display *dpy)
{
	ggi_graphtype gmode = GT_TEXT16;
	int magx = 1, magy = 1;
	int DCLKdiv = 8;
	int dotclock;
	struct kgi_timing tx,ty;
	save_state(&reset_state);

	if (reset_state.ATC[0x10] & AR10_GRAPHICS) {
		switch (reset_state.GRC[0x05] & GR5_WRITEMODE3) {
			case GR5_WRITEMODE0:
			    gmode = GT_8BIT; /* 8 bit mode */
			break;
			case GR5_WRITEMODE1:
			gmode = GT_1BIT; /* 1 bit mode */
			break;
			case GR5_WRITEMODE3:
			gmode = GT_4BIT; /* 4 bit mode */
			break;
			case GR5_WRITEMODE2:
			break;
		}
		magy = (reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	} else { /* Text mode */
		gmode = GT_TEXT16;
		DCLKdiv = (reset_state.SEQ[0x01] & SR01_8DOT_CHCLK) ? 8 : 9;
	}
    
	/* Get horizontal timing */
	tx.width	= (reset_state.CRT[0x01]+1) * DCLKdiv;
	tx.blankstart	= reset_state.CRT[0x02] * DCLKdiv;
	tx.syncstart	= reset_state.CRT[0x04] * DCLKdiv;
	tx.syncend	= ((reset_state.CRT[0x05]) & CR05_HSE_MASK) * DCLKdiv;
	while (tx.syncend < tx.syncstart)
		tx.syncend += 0x20 * DCLKdiv;
	tx.blankend	= ((reset_state.CRT[0x03] & CR03_HBE_MASK) |
			   ((reset_state.CRT[0x05] & CR05_HBLANKEND_B5) ? 0x20 : 0) |
			   (reset_state.CRT[0x01] & 0xC0)) * DCLKdiv;
	tx.total	= (reset_state.CRT[0x00]+5) * DCLKdiv;
	tx.polarity	= (reset_state.MISC & MISC_NEG_HSYNC) != 0;

	/* Get vertical timing */
	ty.width	= 1 + (reset_state.CRT[0x12] | 
			       ((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B8) ? 0x100 : 0) |
			       ((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B9) ? 0x200 : 0));
	ty.blankstart	= 1 + (reset_state.CRT[0x15] |
			       ((reset_state.CRT[0x07] & CR07_VBLANKSTART_B8) ? 0x100 : 0) |
			       ((reset_state.CRT[0x09] & CR09_VBLANKSTART_B9) ? 0x200 : 0));
	ty.syncstart	= (reset_state.CRT[0x10] |                
			   ((reset_state.CRT[0x07] & CR07_VSYNCSTART_B8) ? 0x100 : 0) |
			   ((reset_state.CRT[0x07] & CR07_VSYNCSTART_B9) ? 0x200 : 0));
	ty.syncend	= (reset_state.CRT[0x11] & CR11_VSYNCEND_MASK) |
                (reset_state.CRT[0x12] & ~CR11_VSYNCEND_MASK);
	while (ty.syncend < ty.syncstart)
		ty.syncend += 0x10;
	ty.blankend	= (reset_state.CRT[0x16] | ((ty.width - 1) & ~0xFF)) + 1;
	ty.total = 2 + (reset_state.CRT[0x06] |
			((reset_state.CRT[0x07] & CR07_VTOTAL_B8) ? 0x100 : 0) |
			((reset_state.CRT[0x07] & CR07_VTOTAL_B9) ? 0x200 : 0));
	ty.polarity = (reset_state.MISC & MISC_NEG_VSYNC) != 0;
	
	dotclock = (getCLK() == MISC_25MHZ_CLK) ? 25175000 : 28322000;
   
	/* Set all the mode info that we've just gathered */
	default_ggi_mode.frames		= 1;
	default_ggi_mode.visible.x	= tx.width;
	default_ggi_mode.visible.y	= ty.width;
	default_ggi_mode.virtual.x	= tx.width;
	default_ggi_mode.virtual.y	= ty.width;
	default_ggi_mode.size.x		= 0;
	default_ggi_mode.size.y		= 0;
	default_ggi_mode.graphtype	= gmode;
	default_ggi_mode.text.x		= DCLKdiv;
	default_ggi_mode.text.y		= (reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	if (dpy->defmode == NULL)
		dpy->defmode = &default_ggi_mode;
	else
		(*dpy->defmode) = default_ggi_mode;

	dpy->mode.request	= default_ggi_mode;
	dpy->mode.x		= tx;
	dpy->mode.y		= ty;
	dpy->mode.magnify.x	= magx;
	dpy->mode.magnify.y	= magy;
	dpy->mode.dclk		= dotclock;
	dpy->mode.clk.mul	= 1;
	dpy->mode.clk.div	= 1;
	dpy->mode.lclk.mul	= 1;
	dpy->mode.lclk.div	= 1;
	dpy->mode.chipset	= 0;
	
	default_kgi_mode = dpy->mode;

	chipset_setup_mode(&default_kgi_mode);
	chipset_setup_dpy(dpy, &(default_kgi_mode));

	return EOK;
}
   
static int chipset_init(struct kgi_display *dpy)
{
	int err;
        uint8 result8;
    
	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	if (verify_et6000() == EOK) {
		if ((err = io_check_region(&VGA_IO_REGION)))
			return err;
	        if ((err = io_check_region(&PCI_IO_REGION)))
                if ((err = mem_check_region(&LINEAR_MEM_REGION)))
			return err;

	    mem_claim_region(&LINEAR_MEM_REGION);
 	    io_claim_region(&VGA_IO_REGION);
	    io_claim_region(&PCI_IO_REGION);	    
	    /* Now that we've commandeered PCI I/O region, we can stop using the BIOS calls. */
	    result8 = PCI_in8(dpy, PCI_MMCONTROL);
	    PCI_out8(dpy, result8 | PCI40_LINEARMAP | PCI40_MEMRELOC | PCI40_MEMMAPREG, PCI_MMCONTROL);
	    text16_data_init();
	    return chipset_get_reset_mode(dpy);
	}

	return -E(CHIPSET,FAILED);
}

static void chipset_reset(struct kgi_display *dpy)
{
	restore_state(&reset_state);
	if (dpy) {
		dpy->mode = default_kgi_mode;
		(*dpy->defmode) = default_ggi_mode;
	}
}

static void chipset_done(void)
{
	chipset_reset(NULL);

	io_free_region(&VGA_IO_REGION);
        io_free_region(&PCI_IO_REGION);
        mem_free_region(&et6000_linear_mem);	
}

/* I'm having problems at high refresh with screen distortion; I increased horizontal
 * total _in the monitor driver_ for now. */
static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == GT_TEXT16)
		? mode->request.text.x : 8;

#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
	if (mode->x.total % foo)
		mode->x.total += foo - (mode->x.total % foo);

	if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
		mode->x.syncend -= foo;
	if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
		mode->x.blankend -= foo;

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
		mode->y.syncend--;
	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
		mode->y.blankend--;

#undef ADJUST
}

static int chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
                              enum kgi_tm_cmd cmd)
{
	int bpp = 0;
	int textx, texty;
	textx = mode->request.text.x;
	texty = mode->request.text.y;

	switch (mode->request.graphtype) {
		case GT_8BIT:
		bpp = 8;	/* 1 Bpp */
		break;
	case GT_15BIT:
	case GT_16BIT:
	    bpp = 16;
	    break;
	    
	case GT_24BIT:
	    bpp = 24;
	    break;
	case GT_32BIT:
	    bpp = 32;
	    break;
	    
		case GT_1BIT:
		case GT_4BIT:
		bpp = 4;	/* 0.5 Bpp */
		break;	
		case GT_TEXT16:
		break;		/* Not used in textmode */
		default:
		DEBUG0("graphtype %i not supported", mode->request.graphtype);
		return -E(CHIPSET,FAILED);
	}

#define XVIRTUAL (mode->request.virtual.x)
#define YVIRTUAL (mode->request.virtual.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

	switch (cmd) {
		case CMD_PROPOSE:
		if (mode->request.graphtype == GT_TEXT16) {

			mode->magnify.x = mode->magnify.y = 1;
			if ((textx < 8) || (9 < textx)) {
				DEBUG0("Invalid char width (%i)", textx);
				return -EINVAL;
			}

			mode->request.visible.x -=
				(mode->request.visible.x % textx);
			mode->request.visible.y -=
				(mode->request.visible.y % texty);

			if (XVIRTUAL == 0 && YVIRTUAL == 0)
				return -EINVAL;

			if (XVIRTUAL == 0) 
				XVIRTUAL = TEXT_LEN / ((YVIRTUAL / texty) * 2) * textx;
			
			if (YVIRTUAL == 0)
				YVIRTUAL = TEXT_LEN / ((XVIRTUAL / textx) * 2) * texty;

			if ((2 * (XVIRTUAL / textx)
			     * (YVIRTUAL / texty)) > TEXT_LEN) {
				DEBUG1("Not enough text memory");
				return -ENOMEM;
			}
		} else {	/* mode->request.graphtype == GT_TEXT16 */

			if (XVIRTUAL == 0 && YVIRTUAL == 0)
		             return -EINVAL;

			if (XVIRTUAL == 0)
				XVIRTUAL =  LINEAR_MEM_SIZE / YVIRTUAL;

			if (YVIRTUAL == 0)
				YVIRTUAL =  LINEAR_MEM_SIZE / XVIRTUAL;

			mode->magnify.x = 1;
			mode->magnify.y = 1;
#warning Magnify code is not perfect.
			while ((XVISIBLE * mode->magnify.x) < 400)
				mode->magnify.x++;

			while ((YVISIBLE * mode->magnify.y) < 350) 
				mode->magnify.y++;
			
		        /* Although we can't zoom more than twice AFAIK, the ET6000 PLL can handle
			 * low clocks so an error need not be returned if we try to magnify more
			 * than 2 times.  I'm only magnifying to keep the clocks reasonable any
			 * prevent recentering anyway; it's not necessary. */

		        /* Can't zoom more than this */
			/*if (mode->magnify.y > 2)
		        {
		            return -EINVAL;
		        }*/
		    		
			/* Can't do more than doublescan */
			/*if (mode->magnify.x > 2)
			{
			    return -EINVAL;
			}*/
		    
			if ((textx < 8) || (9 < textx)) {
				DEBUG1("Invalid char width (%i)", textx);
				return -EINVAL;
			}
			if ((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize) {
				if ((XVISIBLE * YVISIBLE * bpp / 8) > chipinfo.ramsize) {
					DEBUG1("Not enough memory to do virtual %ix%i at %i bpp",
					       XVIRTUAL, YVIRTUAL, bpp);
					return -ENOMEM;
				} 
				XVIRTUAL = XVISIBLE;
				YVIRTUAL = YVISIBLE;
			}
			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;
		}

		mode->x.width = mode->request.visible.x * mode->magnify.x;
		mode->y.width = mode->request.visible.y * mode->magnify.y;

		return EOK;
		
		case CMD_LOWER:

		if (mode->dclk < chipinfo.dclk.min) {
			DEBUG1("DCLK too low (%i)", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
		mode->dclk = (mode->dclk > chipinfo.dclk.max)
			? chipinfo.dclk.max : mode->dclk;

		adjust_timing(mode);
		return EOK;

		case CMD_RAISE:

		if (mode->dclk > chipinfo.dclk.max) {
			DEBUG1("DCLK too high (%i)", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
		mode->dclk = (mode->dclk < chipinfo.dclk.min)
			? chipinfo.dclk.min : mode->dclk;

		adjust_timing(mode);
		return EOK;

		case CMD_CHECK:
		if (mode->lclk.mul != 1) {
			DEBUG1("Invalid lclk multiplier %d", mode->lclk.mul);
			return -E(CHIPSET,FAILED);
		}
		switch (mode->lclk.div) {
			case 1:
			break;
			default:
			DEBUG1("Invalid CLK to LCLK divider %d", mode->lclk.div);
			return -E(CHIPSET,FAILED);
	    }
		if (mode->request.graphtype == GT_TEXT16) {
			mode->flags |= FL_CHIP_COLORTEXT;
			if ((mode->magnify.x != 1) || (mode->magnify.y != 1)
			    || (textx < 8) || (9 < textx) || (texty < 1)
			    || (texty > 32) || 
			    ((2 * (XVIRTUAL / textx) * 
			      (YVIRTUAL / texty)) > TEXT_LEN)
			    || (mode->dclk > chipinfo.dclk.max)
			    || (mode->dclk < chipinfo.dclk.min))
			{
				DEBUG1("HW check failed");
				return -E(CHIPSET,FAILED);
			}
		    
		} else {
		    /* what's the point of this... */
		    switch (XVIRTUAL) {
				case 320:
				case 360:
			  case 400:
			  case 512:
				case 640:
				case 720:
			        case 800:
			        case 1024:
			        case 1152:
			        case 1280:
			    
				break;
				default:
			        /*printk("virt check failed\n");
			
				return -E(CHIPSET,FAILED);*/
			}

			if (((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize)
			    || (mode->magnify.x > 2) || (mode->magnify.y > 2)
			    || (mode->clk.mul != 1) || (mode->clk.div != 1)
			    || (mode->dclk < chipinfo.dclk.min)
			    || (chipinfo.dclk.max < mode->dclk)
			    /* LOOKHERE -- allow double pels in > 8bpp modes -- not good code */
			    /*|| (mode->magnify.x == 2 && mode->request.graphtype != GT_8BIT) */ )
		    {
			        NOTICE1("HW check failed\n");
				return -E(CHIPSET,FAILED);
		    }
		    
		}

#define TM(X) ((chipinfo.maxdots.x < mode->x.X) || (chipinfo.maxdots.y < mode->y.X))
		if (TM(width) || TM(blankstart) || TM(syncstart)
		    || TM(syncend) || TM(blankend) || TM(total)) {
			NOTICE1("Timing check failed");
			DEBUG1("Maxdots are (%d,%d)", chipinfo.maxdots.x,
				chipinfo.maxdots.y);
			return -E(CHIPSET,FAILED);
		}
#undef TM

		chipset_setup_mode(mode);
	        adjust_timing(mode);
	    
	    return EOK;

		default:
  	        INTERNALERROR;
		return -E(CHIPSET,FAILED);
	}
	
#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
}

static void chipset_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
	struct statetable setup = default_state;
	register int foo;
	int DCLKdiv;
	int textx,texty;
        struct statetable sane_state;
    
	textx = mode->request.text.x;
	texty = mode->request.text.y;

	/* Set up bits per pixel and logical screen width */

	switch (mode->request.graphtype) {
		case GT_TEXT16:
		foo = 2 * mode->request.virtual.x / textx;
		foo /= 4;	/* Due to the addressing mode used */
		break;
		case GT_1BIT:
		case GT_4BIT:
		    foo = mode->request.virtual.x >> 4;
		break;
		case GT_8BIT:
		    foo = mode->request.virtual.x >> 3;
       	            break;
		case GT_15BIT:
		case GT_16BIT:
		    foo = mode->request.virtual.x >> 2;
	            break;
		case GT_24BIT:
		    foo = (mode->request.virtual.x >> 2) + mode->request.virtual.x;
	            break;
		case GT_32BIT:
		    foo = mode->request.virtual.x >> 1;
		    break;
		default:
		INTERNALERROR;
		return;
	}

	setup.CRT[CRT_LOGICALWIDTH] |= foo & 0xFF;
       	setup.CRT[CRT_OVRFLW_H] |= (foo & 0x100) ? CR3F_LOGICALWIDTH_B8 : 0;
	setup.CRT[CRT_OVRFLW_H] |= (foo & 0x200) ? CR3F_LOGICALWIDTH_B9 : 0;
    
	if (mode->request.graphtype == GT_TEXT16) {
		setup.CRT[0x17] = CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC | 
			CR17_CGA_BANKING | CR17_HGC_BANKING;

		setup.CRT[0x14] |= CR14_UNDERLINE_MASK;

		setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;

		setup.GRC[0x05] |= GR5_ODD_EVEN;
		setup.GRC[0x06] |= GR6_CHAIN_OE;
	        setup.GRC[0x06] |= GR6_MAP_B8_32;

		switch ((DCLKdiv = textx)) {
			case 8:
			setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
			break;
			case 9:
			setup.ATC[0x10] |= AR10_9DOT_CHARS;
			break;
			default:
			INTERNALERROR;
			return;
		}
		setup.CRT[0x09] |= texty - 1;
		setup.ATC[0x13] = (textx == 9) ? 0x08 : 0x00;
	} else {	/* mode->request.graphtype == GT_TEXT16 */
		DCLKdiv = 8;

		setup.CRT[0x17] = CR17_CGA_BANKING | CR17_HGC_BANKING
			| CR17_ENABLE_SYNC | CR17_CGA_ADDR_WRAP;

		for (foo = 0; foo < 16; foo++)
			setup.ATC[foo] = foo;

		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
		setup.SEQ[0x02] |= SR02_PLANE_MASK;
		setup.SEQ[0x04] |= SR04_256K_ACCESS | SR04_NO_ODDEVEN;

		setup.ATC[0x10] |= AR10_GRAPHICS | AR10_TOP_PAN;

		setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64;

		setup.CRT[0x03] |= CR03_IS_VGA;
		setup.CRT[0x09] |= (mode->magnify.y - 1) & CR09_CHARHEIGHT_MASK;

		switch (mode->request.graphtype) {
		        case GT_1BIT:
			case GT_4BIT:
			if (mode->request.graphtype == GT_1BIT) {
				setup.SEQ[0x02] = SR02_PLANE_MASK;
				setup.GRC[0x01] = 0x0e;	/* Set/Reset all but plane 0 */
				setup.GRC[0x04] = 0;
				setup.GRC[0x05] |= GR5_WRITEMODE0;
			} else {
				setup.GRC[0x01] = 0x0f; /* Set/Reset all planes */
				setup.GRC[0x05] |= GR5_WRITEMODE3;
			}

			setup.CRT[0x17] |= CR17_BYTE_MODE;
			break;
		case GT_8BIT:
		case GT_15BIT:
		case GT_16BIT:
		case GT_24BIT:
		case GT_32BIT:
		    
		    setup.SEQ[0x02] |= 0x0f;
		    setup.SEQ[0x04] |= SR04_CHAINED;
		    setup.CRT[0x14] |= CR14_DOUBLEWORDMODE;
		    
		    /* Double the pixels for low-res modes, just like VGA.  This
		     * is not strictly necessary as the ET6000 clocks can handle
		     * frequencies of 12.75 Mhz (320x200), whereas the VGA uses 25 Mhz
		     * and sets this ATC bit. It may be better to remove this, unless
		     * low dot clocks make you uncomfortable.  Experimental. (As is the
                     * rest of the driver!) */
									      
		    if (mode->magnify.x == 2)
		    {
			
			setup.ATC[0x10] |= AR10_8BIT_MODE;
		    }
		    else
		    {
			setup.CRT[0x17] |= CR17_WORDMODE; /* inc memory addr only every 2 cycles */
			setup.CRT[0x14] |= CR14_MEMCOUNT_BY_4;
		    }
		    
		    setup.ATC[0x11] =  0x00;    /* overscan */
		    setup.GRC[0x05] |= GR5_8BIT_COLOR;
		    setup.GRC[0x08] |= 0xff;	/* Enable all bits */
		     
		    break;
		default:
		    INTERNALERROR;
		    return;
		}
		/* note !! fix this, by introducing bpp variable */
	        /* It's unlikely, but PCIVidCtrl may already contain bits which could
		 * be overwritten with zeros.  I will investigate.  */
		switch (mode->request.graphtype) {
		   case GT_15BIT:
		       setup.PCIVidCtrl1 &= ~PCI58_BPP16;
		       /*PCI_out8(dpy, PCI_in8(dpy, 0x58) & ~0x02, 0x58);*/
		       setup.ATC[0x16] |= (1 << 4) | 0x80;
		       break;
  		   case GT_16BIT:
		       setup.ATC[0x16] |= (1 << 4) | 0x80;
		       setup.PCIVidCtrl1 |= PCI58_BPP16;
		       break;
		   case GT_24BIT:
		       setup.ATC[0x16] |= (2 << 4) | 0x80;
		       break;
		   case GT_32BIT:
		       setup.ATC[0x16] |= (3 << 4) | 0x80;
		       break;
		  default:
		}
	}

	/* Set up horizontal timing */
	foo = (mode->x.width / DCLKdiv) - 1;	/* Horizontal display end */
	setup.CRT[0x01] |= foo & 0xFF;
	/*setup.CRT[CRT_OVRFLW_H] |= (foo & 0x100) ? CR3F_HDISPLAYEND_B8 : 0;*/
   
	foo = (mode->x.blankstart / DCLKdiv);	/* Horizontal blank start */
	setup.CRT[0x02] |= foo & 0xFF;
    	setup.CRT[0x3F] |= (foo & 0x100) ? CR3F_HBLANKSTART_B8 : 0;

    
	foo = (mode->x.syncstart / DCLKdiv);	/* Horizontal sync start  */
	setup.CRT[0x04] |= foo & 0xFF;
	setup.CRT[0x3F] |= (foo & 0x100) ? CR3F_HSYNCSTART_B8 : 0;
    
	foo = (mode->x.syncend / DCLKdiv);	/* Horizontal sync end    */
	setup.CRT[0x05] |= foo & CR05_HSE_MASK;
     
	foo = (mode->x.blankend / DCLKdiv);	/* Horizontal blank end   */
	setup.CRT[0x03] |= foo & CR03_HBE_MASK;
	setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
    
	/* Think about adding a little to this if it's the card and not the monitor. */
        /* The problem may be Jaton's card and not the ET6000. */
        foo = (mode->x.total / DCLKdiv) - 5;	/* Horizontal total       */
	setup.CRT[0x00] |= foo & 0xFF;
	setup.CRT[0x3F] |= (foo & 0x100) ? CR3F_HTOTAL_B8 : 0;

	/* Set up vertical timing */

	foo = mode->y.width - 1;		/* Vertical display end   */
	setup.CRT[0x12] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VDISPLAYEND_B10 : 0;
    
	foo = mode->y.blankstart - 1;		/* Vertical blank start   */
	setup.CRT[0x15] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8 : 0;
	setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VBLANKSTART_B10 : 0;

	foo = mode->y.syncstart;		/* Vertical sync start    */
	setup.CRT[0x10] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VSYNCSTART_B10 : 0;

	foo = mode->y.syncend;			/* Vertical sync end      */
	setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;

	foo = mode->y.blankend + 1;		/* Vertical blank end     */
	setup.CRT[0x16] |= foo & 0xFF;

	foo = mode->y.total - 2;		/* Vertical total         */
	setup.CRT[0x06] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VTOTAL_B10 : 0;
    
	setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
	setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

	/* Set splitline to 0 */

	foo = 0;
	setup.CRT[0x18] |= foo & 0xFF;

	setup.CRT[0x07] |= (foo & 0x0100) ? CR07_LINECOMPARE_B8 : 0;
	setup.CRT[0x09] |= (foo & 0x0200) ? CR09_LINECOMPARE_B9 : 0;
    	setup.CRT[0x35] |= (foo & 0x0400) ? CR35_LINECOMPARE_B10 : 0;
    
	/* Ok.  We'll do the actual hardware programming now. */
    
	kgim_monitor_set_mode(dpy, mode);
	kgim_ramdac_set_mode(dpy, mode);
	kgim_clock_set_mode(dpy, mode);
    
        restore_state(&setup);
	     
#ifdef DEBUG_STATE
        printk("*** Dumping new mode\n");
        save_state(&sane_state);
        dump_state(&sane_state);
#endif    
    
	chipset_setup_dpy(dpy, mode);
        
	DEBUG2("Mode set complete\n");
}

static int chipset_command(struct kgi_graphic *state,uint cmd,uint arg)
{
	ggi_coord  *coord;
	ggi_font *font;
	struct kgi_display *dpy;

	dpy = state->dev.dpy;
	switch (cmd) {
		case CHIP_SETDEFFONT:
		font = (struct kgi_font *)(state->io_buf);
	        text16_set_font(dpy, font, 0, font->info->positions);
		return 0;
		case CHIP_SETSPLITLINE:
			if (arg > state->dev.mode.virtual.y)
	                {
				return -E(CHIPSET,BOGUS);
			}
	    
			kgi_set_split(&(state->dev),arg*dpy->mode.magnify.y);
			return 0;
		case CHIP_SETVISFRAME:
			coord = (ggi_coord *)(state->io_buf);

			if (coord->x < 0 || 
			    coord->x > state->dev.mode.virtual.x)
				return -E(CHIPSET,BOGUS);

	    	        if (coord->y < 0 || 
		            coord->y > state->dev.mode.virtual.y)
		            return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
		default:
			break;
	}
	return -E(CHIPSET,UNKNOWN);
}

int kgim_chipset_init(struct kgi_card *card,struct kgi_display *dpy)
{
	int err;

    /* Not used. */
    /*private = kgi_malloc_small(sizeof(*private), GFP_KERNEL);
    if (private == NULL)
    {
	NOTICE2("ERROR: failed to allocate private data!");
	return -ENOMEM;
    }
    memset(private, 0, sizeof(*private));
    KGIM_PRIVATE(dpy, chipset) = private;   
    */
	if ((err = chipset_init(dpy)) != EOK) return err;
 
        return err;
	strcpy(card->manufact, chipinfo.manufact);
	strcpy(card->model, chipinfo.model);
	card->modes = chipinfo.modes;
	card->flags = chipinfo.flags;
	card->maxdots = chipinfo.maxdots;

	card->chipset =& chipinfo;
	card->preset = default_kgi_mode;

	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	chipset_done();
}

int  kgim_chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			     enum kgi_tm_cmd cmd)
{
	int val;
        DEBUG1("Checking mode... (%dx%d %d)\n", mode->request.virtual.x,
	       mode->request.virtual.y, mode->request.graphtype);
	val = chipset_check_mode(dpy, mode, cmd);
        DEBUG2("Mode check %s\n", (val == EOK) ? "succeeded" : "failed");
    return val;
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
        DEBUG1("Setting mode... (%dx%d %d)\n", mode->request.virtual.x,
	mode->request.virtual.y, mode->request.graphtype);
	chipset_set_mode(dpy,mode);
}

int kgim_chipset_command(struct kgi_graphic *state, uint cmd, uint iobuf)
{
	return chipset_command(state, cmd, iobuf);
}
