/*\
|*| --------------------------------------------------------------------------
|*|	ET6000 I/O
|*| --------------------------------------------------------------------------
|*|
|*|     Copyright (C) 1997 by Jim Ursetto,    [ursetto@uiuc.edu]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|	This code was derived from the following source of information:
|*|	[1]	ET4000 driver (c) Jason McMullan and Andreas Vogler
|*|	[2]	Prerelease VGADOC5 by Finn Thoegersen
\*/

/*\
|*|     GGI includes
\*/

#include "IBM/vga_io.inc"

/*\
|*|	generalized DAC output functions
\*/

static void enable_ext(void)
{
   io_out8(0x03, HERC_CMPT);   /* from et6000.h */
   io_out8(0xa0, DISPLAY_MODE);
}

static void disable_ext(void)
{
    io_out8(0x29, DISPLAY_MODE);
    io_out8(0x01, HERC_CMPT);
}

static unsigned short dac_port[4] ={ 0x3C8, 0x3C9, 0x3C6, 0x3C7 };

void DAC_out8(struct kgi_display *dpy,uint8 value, uint8 dac_reg)
{
    if (dac_reg & ~0x3)
      printk("%s: warning: DAC register select lines truncated: %x to %x.\n",\
	     __FILE__, dac_reg, dac_reg &= 0x3);
    
    io_out8(value, dac_port[dac_reg & 0x03]);
}

uint8 DAC_in8(struct kgi_display *dpy,uint8 dac_reg)
{
    if(dac_reg & ~0x03) 
      printk("%s: warning: DAC register select lines truncated: %x to %x.\n",\
	     __FILE__, dac_reg, dac_reg &= 0x03);
    
    return io_in8(dac_port[dac_reg & 0x03]);
}

void DAC_ins8(struct kgi_display *dpy,uint8 dac_reg,void *buf, uint count)
{
    if(dac_reg & ~0x03)
      printk("%s: warning; DAC register select lines truncated: %x to %x.\n",\
	     __FILE__, dac_reg, dac_reg &= 0x3);
    
    io_ins8(dac_port[dac_reg & 0x03], buf, count);
}

void DAC_outs8(struct kgi_display *dpy,uint8 dac_reg, void *buf, uint count)
{
    if(dac_reg & ~0x03) 
      printk("%s: warning: DAC register select lines truncated: %x to %x.\n",\
	     __FILE__, dac_reg, dac_reg &= 0x03);
    
    io_outs8(dac_port[dac_reg & 0x03], buf, count);
}

/*\
|*|	clock I/O functions
\*/

/* Selects a specific clock.  Doesn't change value of clock timing! */
void CLK_out8(struct kgi_display *dpy,uint8 value,uint8 reg)
{
    if (reg!=0)
      INTERNALERROR;
    PASSERT(!(value & ~0x07), "trying to select clock: %x", value);
    DEBUG1("selecting clock #%d\n", value);
    enable_ext();
    outbMISC((io_in8(MISCr) & ~MISC_CLOCK_MASK) | ((value & 0x03) << 2));
    outbCRT((inbCRT(CRT_CCR) & ~CR34_CLK_SEL2) | ((value & 0x04) >> 1),
	    CRT_CCR);
    /* ignore high bits */
    outbCRT((inbCRT(CRT_GENERAL) & ~CR31_CLK_MASK), CRT_GENERAL);
    DEBUG2("selected clock #%d\n", ((io_in8(MISCr) & MISC_CLOCK_MASK) >> 2)
                          | ((inbCRT(CRT_CCR) & CR34_CLK_SEL2) << 1));
    disable_ext();			  
}

uint8 CLK_in8(struct kgi_display *dpy,uint8 reg)
{
    uint8 ret=0;
    if (reg!=0)
      INTERNALERROR;
    
    ret |= (io_in8(MISCr) & MISC_CLOCK_MASK) >> 2;
    ret |= (inbCRT(CRT_CCR) & CR34_CLK_SEL2) << 1;
    /* It is unknown whether or not index 31 of the CRT register
     * should be returned. VGADOC is unclear. */
    
    return ret;
}

#define NoPCIRegs 0x87

/* These are non-static !! */

void PCI_out8(struct kgi_display *dpy, uint8 value, uint8 reg)
{
    ASSERT(reg < NoPCIRegs);
    io_out8(value,PCI_IO+reg);
}

uint8 PCI_in8(struct kgi_display *dpy, uint8 reg)
{
    ASSERT(reg < NoPCIRegs);
    return io_in8(PCI_IO+reg);
}

#define NoSEQRegs	0x08
#define	NoCRTRegs	0x40
#define NoGRCRegs	0x09
#define NoATCRegs	0x18

struct statetable
{
    uint8 MISC;
    uint8 SEQ[NoSEQRegs];
    uint8 CRT[NoCRTRegs];
    uint8 GRC[NoGRCRegs];
    uint8 ATC[NoATCRegs];
    uint8 PCIVidCtrl1;
};

static void save_SEQ(uint8 *buf)
{
    int i;
    
    for(i = 0; i < 5; i++) 
      *(buf++)=inbSEQ(i);
    buf++;
    for(i = 6; i < 8; i++) 
      *(buf++)=inbSEQ(i);
}

static void restore_SEQ(uint8 *buf)
{
    register uint8 i;
    for(i=0; i < 5; i++) 
      outbSEQ(*(buf++),i);
    buf++;
    for(i=6; i < 8; i++) 
      outbSEQ(*(buf++),i);
}

static void save_CRT(uint8 *buf)
{
    int i;
    
    for(i = 0; i < 0x19; i++) 
      *(buf++)=inbCRT(i);
    buf+=0x17;
    for(i = 0x30; i < 0x40; i++) 
      *(buf++)=inbCRT(i);
}

static void restore_CRT(uint8 *buf)
{
    register uint8 i;

    outbCRT(inbCRT(CRT_VSYNCEND) & ~CR11_LOCKTIMING, CRT_VSYNCEND);
    outbCRT(inbCRT(CRT_HBLANKEND) | CR03_IS_VGA,CRT_HBLANKEND);
    
    for(i = 0; i < 0x19; i++) {
	switch (i) {
	case CRT_HBLANKEND:
	    *buf |= CR03_IS_VGA;
	    break;
	case CRT_VSYNCEND:
	    *buf &= ~CR11_LOCKTIMING;
	    break;
	default:
	    break;
	}
	outbCRT(*(buf++), i);
    }
    buf+=0x17;

    for(i = 0x30; i < 0x40; i++) {
	switch(i) {

       /*case CRT_GENERAL:
	      *buf = (io_in8(CRT_GENERAL) & CR31_CLK_MASK)
	         | (*buf & ~CR31_CLK_MASK);
	    break;*/
	case CRT_CCR:
	    *buf = (inbCRT(CRT_CCR) & CR34_CLK_SEL2) 
	           | (*buf & ~CR34_CLK_SEL2);
	    break;
	default:
	    break;
	}
	outbCRT(*(buf++), i);
    }

}

static void save_GRC(uint8 *buf)
{
    int i;
    
    for(i = 0; i < NoGRCRegs; i++) 
      *(buf++)=inbGRC(i);
}

static void restore_GRC(uint8 *buf)
{
    register uint8 i;
    
    for(i = 0; i < NoGRCRegs; i++) 
      outbGRC(*(buf++),i);
}

static void save_ATC(uint8 *buf)
{
    int i;
    
    for(i = 0; i < 0x15; i++) {
	io_in8(INS1r);  /* reset AFF and setup index */
	*(buf++)=inbATC(i);
    }
    buf++;
    for(i = 0x16; i < 0x18; i++) {
	io_in8(INS1r);  
	*(buf++)=inbATC(i);
    }
    io_in8(INS1r); 
    io_out8( ATCI_ENABLEDISPLAY, ATCI);
}

static void restore_ATC(uint8 *buf)
{
    register uint8 i;
    
    for(i = 0; i < 0x15; i++) {
	io_in8(INS1r); 
	outbATC(*(buf++),i);
    }
    buf++;
    for(i = 0x16; i < 0x18; i++) {
	io_in8(INS1r); 
	outbATC(*(buf++),i);
    }
    io_in8(INS1r); 
    io_out8(ATCI_ENABLEDISPLAY, ATCI);
}

static void restore_PCI(struct statetable *state)
{
    io_out8(state->PCIVidCtrl1, PCI_IO+PCI_VIDCTRL1);
}    

static void save_PCI(struct statetable *state)
{
    state->PCIVidCtrl1 = io_in8(PCI_IO+PCI_VIDCTRL1);
}    

static void
save_state(struct statetable *state)
{
    state->MISC  = io_in8(MISCr);	/* save miscellaneous output reg */

    save_SEQ(state->SEQ);	/* save sequencer regs */
    save_CRT(state->CRT);	/* save cathode ray tube controller regs */
    save_GRC(state->GRC);	/* save graphics controller regs */
    save_ATC(state->ATC);	/* save attribute controller regs */
    save_PCI(state);            /* save PCI regs */
}

static void restore_state(struct statetable *state)
{
    /* changed for now since we * want * to restore the clock, I think */
    /* !! Changed back.  restore_state is called as a mode set after the */
    /* clock is set in kgim_clock_set.  da kara, do not reset clock here!! */

    outbMISC((io_in8(MISCr) & MISC_CLOCK_MASK) | state->MISC);
    enable_ext();
    restore_SEQ(state->SEQ);
    restore_CRT(state->CRT);
    restore_GRC(state->GRC);
    restore_ATC(state->ATC);
    restore_PCI(state);
    disable_ext();
}

#if defined(DEBUG_IO)||defined(DEBUG_STATE)
static void dump_state(struct statetable *state)
{ int i;
  printk("MISC: %.2x",state->MISC);
  
  for(i = 0; i < NoSEQRegs; i++)
   { if(!(i & 0x7)) printk("\nSEQ:%.2x: ",i); printk("0x%.2x, ",state->SEQ[i]); }
  for(i = 0; i < NoCRTRegs; i++)
   { if(!(i & 0x7)) printk("\nCRT:%.2x: ",i); printk("0x%.2x, ",state->CRT[i]); }
  for(i = 0; i < NoGRCRegs; i++)
   { if(!(i & 0x7)) printk("\nGRC:%.2x: ",i); printk("0x%.2x, ",state->GRC[i]); }
  for(i = 0; i < NoATCRegs; i++)
   { if(!(i & 0x7)) printk("\nATC:%.2x: ",i); printk("0x%.2x, ",state->ATC[i]); }
  printk("\n");
}
#endif

